/* FIXME: This ‘negative’ variant of a GtkSwitch should probably be
 * upstreamed to GTK. See https://gitlab.gnome.org/GNOME/gtk/issues/2470 */
switch:checked.restricts {
  background-color: @yellow_5;
}

switch:checked.restricts, switch:checked.restricts slider {
  border-color: @yellow_3;
}

switch:disabled.restricts {
  border-color: @borders;
  background-color: @insensitive_bg_color;
}

switch:disabled.restricts slider {
  filter: grayscale(100%);
}
